import torch
from torch.distributions import uniform
from typing import List, Union, Tuple, Optional
from perceiver_pytorch.utils import encode_position
from math import prod
import einops
import logging

_LOG = logging.getLogger("perceiver.queries")
_LOG.setLevel(logging.WARN)


class LearnableQuery(torch.nn.Module):
    """
    Module that constructs a learnable query of query_shape for the Perceiver
    """

    def __init__(
        self,
        channel_dim: int,
        query_shape: Union[Tuple[int], List[int]],
        conv_layer: str = "3d",
        max_frequency: float = 16.0,
        num_frequency_bands: int = 64,
        frequency_base: float = 2.0,
        sine_only: bool = False,
        precomputed_fourier: Optional[torch.Tensor] = None,
        generate_fourier_features: bool = False,
    ):
        """
        Learnable Query with some inbuilt randomness to help with ensembling

        Args:
            channel_dim: Channel dimension for the output of the network
            query_shape: The final shape of the query, generally, the (T, H, W) of the output
            conv_layer: The type of convolutional layer to use, either 3d or 2d
            max_frequency: Max frequency for the Fourier Features
            num_frequency_bands: Number of frequency bands for the Fourier Features
            frequency_base: Base frequency for the Fourier Features
            sine_only: Whether to use only the sine Fourier features
            precomputed_fourier: Fourier features to use instead of computing them here,
                useful for having temporally consistent features from history timesteps to future predictions
                These features will be concatenated directly to the query, so should be compatible, and made in the
                same way as in encode_position
            generate_fourier_features: Whether to use generated Fourier features giving the relative
            position within the predictions.
        """
        super().__init__()
        self.query_shape = query_shape
        self.generate_fourier_features = generate_fourier_features
        # Need to get Fourier Features once and then just append to the output
        fourier_features = []
        if precomputed_fourier is not None:
            fourier_features += [precomputed_fourier]
        if self.generate_fourier_features:
            generated_features = encode_position(
                1,  # Batch size, 1 for this as it will be adapted in forward
                axis=query_shape,
                max_frequency=max_frequency,
                num_frequency_bands=num_frequency_bands,
                sine_only=sine_only,
            )
            fourier_features += [generated_features]
        if len(fourier_features) > 1:
            self.fourier_features = torch.cat(fourier_features, dim=-1)
        elif fourier_features:
            # Only have one of the two options
            self.fourier_features = fourier_features[0]
        else:
            # None are set
            self.fourier_features = None

        self.channel_dim = channel_dim
        if (
            conv_layer == "3d" and len(self.query_shape) == 3
        ):  # If Query shape is for an image, then 3D conv won't work
            conv = torch.nn.Conv3d
        elif conv_layer == "2d":
            conv = torch.nn.Conv2d
        else:
            raise ValueError(f"Value for 'layer' is {conv_layer} which is not one of '3d', '2d'")
        self.conv_layer = conv_layer
        self.layer = conv(
            in_channels=channel_dim, out_channels=channel_dim, kernel_size=3, padding=1
        )
        # Linear layer to compress channels down to query_dim size?
        self.fc = torch.nn.Linear(self.channel_dim, self.channel_dim)
        self.distribution = uniform.Uniform(low=torch.Tensor([0.0]), high=torch.Tensor([1.0]))

    def output_shape(self) -> Tuple[int, int]:
        """
        Gives the output shape from the query, useful for setting the correct
        query_dim in the Perceiver

        Returns:
            The shape of the resulting query, excluding the batch size
        """

        # The shape is the query_dim + Fourier Feature channels
        if self.fourier_features is not None:
            channels = self.fourier_features.shape[-1] + self.channel_dim
        else:
            channels = self.channel_dim
        return prod(self.query_shape), channels

    def forward(
        self, x: torch.Tensor, fourier_features: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        """
        Samples the uniform distribution and creates the query by passing the
        sample through the model and appending Fourier features

        Args:
            x: The input tensor to the model, used to batch the batch size
            fourier_features: Fourier features to append to the input, if used, the output_shape will be incorrect,
                and to get the correct output shape, the fourier_features channels have to be added to output_shape

        Returns:
            Torch tensor used to query the output of the PerceiverIO model
        """
        _LOG.debug(f"Batch: {x.shape[0]} Query: {self.query_shape} Dim: {self.channel_dim}")
        z = self.distribution.sample((x.shape[0], self.channel_dim, *self.query_shape)).type_as(
            x
        )  # [B, Query, T, H, W, 1] or [B, Query, H, W, 1]
        z = torch.squeeze(z, dim=-1)  # Extra 1 for some reason
        _LOG.debug(f"Z: {z.shape}")
        # Do 3D or 2D CNN to keep same spatial size, concat, then linearize
        if self.conv_layer == "2d" and len(self.query_shape) == 3:
            # Iterate through time dimension
            outs = []
            for i in range(x.shape[1]):
                outs.append(self.layer(z[:, :, i, :, :]))
            query = torch.stack(outs, dim=2)
        else:
            query = self.layer(z)
        # Move channels to correct location
        query = einops.rearrange(query, "b c ... -> b ... c")
        to_concat = [query]
        if self.fourier_features is not None:
            ff = einops.repeat(
                self.fourier_features, "b ... -> (repeat b) ...", repeat=x.shape[0]
            )  # Match batches
            to_concat = to_concat + [ff]
        if fourier_features is not None:
            to_concat = to_concat + [fourier_features]
        if len(to_concat) > 1:
            query = torch.cat(to_concat, dim=-1)
        # concat to channels of data and flatten axis
        query = einops.rearrange(query, "b ... d -> b (...) d")
        _LOG.debug(f"Final Query Shape: {query.shape}")
        return query



import torch
from einops import rearrange, repeat
from torch import nn

from perceiver_pytorch.layers import exists, cache_fn, PreNorm, FeedForward, Attention
from perceiver_pytorch.rotary import SinusoidalEmbeddings
from perceiver_pytorch.utils import encode_position

# main class


class Perceiver(nn.Module):
    def __init__(
        self,
        *,
        num_freq_bands,
        depth,
        max_freq,
        input_channels=3,
        input_axis=2,
        num_latents=512,
        latent_dim=512,
        cross_heads=1,
        latent_heads=8,
        cross_dim_head=64,
        latent_dim_head=64,
        num_classes=1000,
        attn_dropout=0.0,
        ff_dropout=0.0,
        weight_tie_layers=False,
        fourier_encode_data=True,
        sine_only: bool = False,
        self_per_cross_attn=1,
        self_attn_rel_pos=True,
    ):
        """
        Perceiver: https://arxiv.org/abs/2103.03206
        The shape of the final attention mechanism will be:
        depth * (cross attention -> self_per_cross_attn * self attention)

        Args:
          num_freq_bands: Number of freq bands, with original value (2 * K + 1)
          depth: Depth of net.
          max_freq: Maximum frequency, hyperparameter depending on how
              fine the data is.
          input_channels: Number of channels for each token of the input.
          input_axis: Number of axes for input data (2 for images, 3 for video)
          num_latents: Number of latents, or induced set points, or centroids.
              Different papers giving it different names.
          latent_dim: Latent dimension.
          cross_heads: Number of heads for cross attention. Paper said 1.
          latent_heads: Number of heads for latent self attention, 8.
          cross_dim_head: Number of dimensions per cross attention head.
          latent_dim_head: Number of dimensions per latent self attention head.
          num_classes: Output number of classes.
          attn_dropout: Attention dropout
          ff_dropout: Feedforward dropout
          weight_tie_layers: Whether to weight tie layers (optional).
          fourier_encode_data: Whether to auto-fourier encode the data, using
              the input_axis given. defaults to True, but can be turned off
              if you are fourier encoding the data yourself.
            sine_only: Use only sine encoding in fourier encoding, compared to using sine and cos
          self_per_cross_attn: Number of self attention blocks per cross attn.
          self_attn_rel_pos:
        """
        super().__init__()
        self.input_axis = input_axis
        self.max_freq = max_freq
        self.num_freq_bands = num_freq_bands

        self.fourier_encode_data = fourier_encode_data
        fourier_channels = (input_axis * ((num_freq_bands * 2) + 1)) if fourier_encode_data else 0
        self.sine_only = sine_only
        input_dim = fourier_channels + input_channels

        # Randomly initialise the 'latent array'.
        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))

        def get_cross_attn():
            return PreNorm(
                latent_dim,
                Attention(
                    latent_dim,
                    input_dim,
                    heads=cross_heads,
                    dim_head=cross_dim_head,
                    dropout=attn_dropout,
                ),
                context_dim=input_dim,
            )

        def get_cross_ff():
            return PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))

        def get_latent_attn():
            return PreNorm(
                latent_dim,
                Attention(
                    latent_dim,
                    heads=latent_heads,
                    dim_head=latent_dim_head,
                    dropout=attn_dropout,
                ),
            )

        def get_latent_ff():
            return PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))

        # Cache all the above functions.
        get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff = map(
            cache_fn, (get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff)
        )

        self.layers = nn.ModuleList([])
        for i in range(depth):
            should_cache = i > 0 and weight_tie_layers
            cache_args = {"_cache": should_cache}

            self_attns = nn.ModuleList([])

            for _ in range(self_per_cross_attn):
                self_attns.append(
                    nn.ModuleList(
                        [
                            get_latent_attn(**cache_args),
                            get_latent_ff(**cache_args),
                        ]
                    )
                )

            self.layers.append(
                nn.ModuleList(
                    [
                        get_cross_attn(**cache_args),
                        get_cross_ff(**cache_args),
                        self_attns,
                    ]
                )
            )

        self.to_logits = nn.Sequential(nn.LayerNorm(latent_dim), nn.Linear(latent_dim, num_classes))

        self.sinu_emb = None
        if self_attn_rel_pos:
            self.sinu_emb = SinusoidalEmbeddings(latent_dim_head)

    def forward(self, data, mask=None):
        """
        Args:
          data: If sequential is True, then data must be of shape:
              (batch size, sequence length, *axes) where axes would be width
              and height for images.
        """

        b, *axis, _ = data.shape
        device = data.device

        assert (
            len(axis) == self.input_axis
        ), f"Input data must have {self.input_axis} axes, not {len(axis)}!"

        if self.fourier_encode_data:
            # Calculate Fourier encoded positions in the range of [-1, 1],
            # for all axes.
            enc_pos = encode_position(
                b,
                axis,
                self.max_freq,
                self.num_freq_bands,
                sine_only=self.sine_only,
            ).type_as(data)

            data = torch.cat((data, enc_pos), dim=-1)

        # Concat to channels of data and flatten axes.
        # b = batch size; d = last dimension of data
        data = rearrange(data, "b ... d -> b (...) d", b=b)

        # x is the 'latent array' in the paper.
        # b = batch size; n = number of latents; d = latent dimensions.
        x = repeat(self.latents, "n d -> b n d", b=b)

        # Rotary embeddings for latents, if specified.
        pos_emb = self.sinu_emb(x) if exists(self.sinu_emb) else None

        # Layers.
        for cross_attn, cross_ff, self_attns in self.layers:
            x = cross_attn(x, context=data, mask=mask) + x
            x = cross_ff(x) + x

            for self_attn, self_ff in self_attns:
                x = self_attn(x, pos_emb=pos_emb) + x
                x = self_ff(x) + x

        x = x.mean(dim=-2)
        return self.to_logits(x)



from perceiver_pytorch.perceiver_io import PerceiverIO
from perceiver_pytorch.modalities import InputModality, modality_encoding
from perceiver_pytorch.utils import encode_position, fourier_encode
import torch
from typing import List, Iterable, Dict, Optional, Any, Union, Tuple
from einops import rearrange, repeat
from math import prod


class MultiPerceiver(torch.nn.Module):
    def __init__(
        self,
        modalities: Iterable[InputModality],
        fourier_encode_data: bool = True,
        input_channels: int = 3,
        output_channels: int = 12,
        forecast_steps: int = 48,
        sine_only: bool = False,
        output_shape: Union[int, Tuple[int, ...]] = 32,
        **kwargs,
    ):
        """
        PerceiverIO made to work more specifically with timeseries images and multimodal inputs https://arxiv.org/abs/2107.14795
        This is a wrapper around the PerceiverIO implementation to encode the inputs correctly

        Args:
            input_channels: Number of input channels (int)
            forecast_steps: Number of forecast steps to make (int)
            fourier_encode_data: Whether to add Fourier Features to the input data, if this is false, inputs should be have some type of positional encoding added beforehand
            output_channels: Number of output channels per image (int)
            sine_only: Only use Sine part of Fourier features (bool)
            output_shape: Int or Tuple of ints, giving the desired output shape of the model
            **kwargs: Extra kwargs to pass through to PerceiverIO
        """
        super(MultiPerceiver, self).__init__()
        self.fourier_encode_data = fourier_encode_data
        self.forecast_steps = forecast_steps
        self.input_channels = input_channels
        self.sine_only = sine_only
        self.output_channels = output_channels
        self.modalities = {modality.name: modality for modality in modalities}
        # we encode modality with one hot encoding, so need one dim per modality:
        modality_encoding_dim = len(modalities)
        # input_dim is the maximum dimension over all input modalities:
        input_dim = max(modality.input_dim for modality in modalities) + modality_encoding_dim
        # Pop dim
        self.max_modality_dim = input_dim
        kwargs.pop("dim", None)
        # Want toe logit_dim to be the same as the channels * width or height
        if isinstance(output_shape, int):
            kwargs["logits_dim"] = output_shape * self.output_channels
        else:
            kwargs["logits_dim"] = prod(output_shape)
        self.perceiver = PerceiverIO(dim=input_dim, **kwargs)

    def decode_output(self, data):
        pass

    def forward(self, multi_modality_data: Dict[str, torch.Tensor], mask=None, queries=None):
        batch_sizes = set()
        num_modalities = len(multi_modality_data)
        linearized_data = []

        for modality_index, modality_name in enumerate(sorted(multi_modality_data.keys())):
            assert (
                modality_name in self.modalities
            ), f"modality {modality_name} was not defined in constructor"
            data = multi_modality_data[modality_name]
            modality = self.modalities[modality_name]
            b, *axis, _ = data.size()
            assert len(axis) == modality.input_axis, (
                f"input data must have the right number of axes for modality {modality_name}. "
                f"Expected {modality.input_axis} while forward argument offered {len(axis)}"
            )
            batch_sizes.add(b)
            assert len(batch_sizes) == 1, "batch size must be the same across all modalities"
            enc_pos = []
            if self.fourier_encode_data:
                # calculate fourier encoded positions in the range of [-1, 1], for all axis
                enc_pos = encode_position(
                    batch_size=b,
                    axis=axis,
                    max_frequency=modality.max_freq,
                    num_frequency_bands=modality.num_freq_bands,
                    sine_only=self.sine_only,
                ).type_as(data)

            # Figure out padding for this modality, given max dimension across all modalities:
            padding_size = self.max_modality_dim - modality.input_dim - num_modalities

            padding = torch.zeros(size=data.size()[0:-1] + (padding_size,)).type_as(data)
            # concat to channels of data and flatten axis
            modality_encodings = modality_encoding(b, axis, modality_index, num_modalities).type_as(
                data
            )
            to_concat = (
                (data, padding, enc_pos, modality_encodings)
                if len(enc_pos) > 0
                else (data, padding, modality_encodings)
            )
            data = torch.cat(to_concat, dim=-1)
            # concat to channels of data and flatten axis
            data = rearrange(data, "b ... d -> b (...) d")
            linearized_data.append(data)

        # Concatenate all the modalities:
        data = torch.cat(linearized_data, dim=1)

        # After this is the PerceiverIO backbone, still would need to decode it back to an image though
        # Should include the query shape here for the output we want, could be learned embeddings, repeated input frames of the same shape that is desired, etc.
        perceiver_output = self.perceiver.forward(data, mask, queries)

        # For multiple modalities, they are split after this back into different tensors
        # For Sat images, we just want the images, not the other ones, so can leave it as is?

        # Have to decode back into future Sat image frames
        # Perceiver for 'pixel' postprocessing does nothing, or undoes the space2depth from before if just image
        # If doing depth2space, should split modalities again

        # Reshape to the correct output
        # This is how it is done in the official implementation, do a decoder query with cross attention, then just reshape the output
        # For timeseries, this is given as a query with T*H*W shape
        # For Flow Decoder, this is the same, except has a rescale factor

        # To keep this more general, leave the reshaping to postprocessing outside the model

        return perceiver_output



import torch
from dataclasses import dataclass


@dataclass
class InputModality:
    name: str
    input_channels: int
    input_axis: int
    num_freq_bands: int
    max_freq: float
    sin_only: bool = False
    fourier_encode: bool = True

    @property
    def input_dim(self) -> int:
        # Calculate the dimension of this modality.
        fourier_channels = self.input_axis * ((self.num_freq_bands * 2) + 1)
        fourier_channels = fourier_channels // 2 if self.sin_only else fourier_channels
        input_dim = fourier_channels + self.input_channels
        return input_dim


def modality_encoding(
    batch_size: int, axes, modality_index: int, num_modalities: int
) -> torch.Tensor:
    """
    Return one-hot encoding of modality given num_modalities, batch size and axes.
    The result need to be compatible with the modality data for concatenation.

    Args:
        batch_size: Batch size of the input
        axes: The size of each axis, other than batch size, of the input
        modality_index:  The index of this modality i.e. if there are 3 modalities, this would be 0, 1, or 2
        num_modalities: Total number of modalities

    Returns:
        One hot encoding of which modality the input is

    """
    one_hot = torch.eye(num_modalities, num_modalities)[modality_index]
    to_expand = [batch_size]
    one_hot = one_hot.unsqueeze(0)
    for i, axis in enumerate(axes):
        one_hot = one_hot.unsqueeze(0)
        to_expand.append(axis)
    to_expand.append(num_modalities)

    one_hot = one_hot.expand(to_expand)
    return one_hot



import torch
from torch import nn, einsum
import torch.nn.functional as F

from einops import rearrange, repeat

from perceiver_pytorch.layers import exists, default, cache_fn, PreNorm, FeedForward, Attention
from perceiver_pytorch.utils import fourier_encode


# latent mixer


def Mixer(seq_len, mult=4, dropout=0.0):
    return nn.Sequential(
        nn.Conv1d(seq_len, seq_len * mult, 1),
        nn.GELU(),
        nn.Dropout(dropout),
        nn.Conv1d(seq_len * mult, seq_len, 1),
    )


# main class


class Perceiver(nn.Module):
    def __init__(
        self,
        *,
        num_freq_bands,
        depth,
        max_freq,
        input_channels=3,
        input_axis=2,
        num_latents=512,
        latent_dim=512,
        cross_heads=1,
        latent_heads=8,
        cross_dim_head=64,
        latent_dim_head=64,
        num_classes=1000,
        attn_dropout=0.0,
        ff_dropout=0.0,
        weight_tie_layers=False,
        **kwargs
    ):
        super().__init__()
        self.input_axis = input_axis
        self.max_freq = max_freq
        self.num_freq_bands = num_freq_bands

        input_dim = input_axis * ((num_freq_bands * 2) + 1) + input_channels

        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))

        get_cross_attn = lambda: PreNorm(
            latent_dim,
            Attention(
                latent_dim,
                input_dim,
                heads=cross_heads,
                dim_head=cross_dim_head,
                dropout=attn_dropout,
            ),
            context_dim=input_dim,
        )
        get_latent_attn = lambda: PreNorm(latent_dim, Mixer(num_latents, dropout=ff_dropout))
        get_cross_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))
        get_latent_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))

        get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff = map(
            cache_fn,
            (get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff),
        )

        self.layers = nn.ModuleList([])
        for i in range(depth):
            should_cache = i > 0 and weight_tie_layers
            cache_args = {"_cache": should_cache}

            self.layers.append(
                nn.ModuleList(
                    [
                        get_cross_attn(**cache_args),
                        get_cross_ff(**cache_args),
                        get_latent_attn(**cache_args),
                        get_latent_ff(**cache_args),
                    ]
                )
            )

        self.to_logits = nn.Sequential(nn.LayerNorm(latent_dim), nn.Linear(latent_dim, num_classes))

    def forward(self, data, mask=None):
        b, *axis, _, device = *data.shape, data.device
        assert len(axis) == self.input_axis, "input data must have the right number of axis"

        # calculate fourier encoded positions in the range of [-1, 1], for all axis

        axis_pos = list(
            map(
                lambda size: torch.linspace(-1.0, 1.0, steps=size, device=device),
                axis,
            )
        )
        pos = torch.stack(torch.meshgrid(*axis_pos), dim=-1)
        enc_pos = fourier_encode(
            x=pos,
            max_freq=self.max_freq,
            num_bands=self.num_freq_bands,
        )
        enc_pos = rearrange(enc_pos, "... n d -> ... (n d)")
        enc_pos = repeat(enc_pos, "... -> b ...", b=b)

        # concat to channels of data and flatten axis

        data = torch.cat((data, enc_pos), dim=-1)
        data = rearrange(data, "b ... d -> b (...) d")

        x = repeat(self.latents, "n d -> b n d", b=b)

        for cross_attn, cross_ff, latent_attn, latent_ff in self.layers:
            x = cross_attn(x, context=data, mask=mask) + x
            x = cross_ff(x) + x
            x = latent_attn(x) + x
            x = latent_ff(x) + x

        x = x.mean(dim=-2)
        return self.to_logits(x)



from functools import wraps

import torch
from einops import rearrange, repeat
from torch import nn, einsum
from torch.nn import functional as F

from perceiver_pytorch.rotary import apply_rotary_emb


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def cache_fn(f):
    cache = None

    @wraps(f)
    def cached_fn(*args, _cache=True, **kwargs):
        if not _cache:
            return f(*args, **kwargs)
        nonlocal cache
        if cache is not None:
            return cache
        cache = f(*args, **kwargs)
        return cache

    return cached_fn


class PreNorm(nn.Module):
    def __init__(self, dim, fn, context_dim=None):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)
        self.norm_context = (
            nn.LayerNorm(context_dim) if exists(context_dim) else None
        )

    def forward(self, x, **kwargs):
        x = self.norm(x)

        if exists(self.norm_context):
            context = kwargs["context"]
            normed_context = self.norm_context(context)
            kwargs.update(context=normed_context)

        return self.fn(x, **kwargs)


class GEGLU(nn.Module):
    """
    Gaussian Error Gated Linear Unit.
    See Shazer 2020: https://arxiv.org/abs/2002.05202
    """
    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


class FeedForward(nn.Module):
    """Feed forward neural net with GEGLU activation."""

    def __init__(self, dim: int, mult: int = 4, dropout: float = 0.0):
        """
        Args:
            dim: Input & Output size.
            mult: The inner dimension of the FF net will be dim * mult.
            dropout: Proportion to dropout after the GEGLU.
        """
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim * mult * 2),
            GEGLU(),
            nn.Dropout(dropout),
            nn.Linear(dim * mult, dim),
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(
        self, query_dim, context_dim=None, heads=8, dim_head=64, dropout=0.0
    ):
        """
        Args:
            query_dim: Size of the queries.
            context_dim: Size of the 'context' (the 'byte array' in the paper).
                If None, will default to the query_dim.
            heads: Number of attention heads.
            dim_head: Number of dimensions per head.
            dropout: Proportion to dropout (in the final linear layer).
        """

        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)

        self.scale = dim_head ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, query_dim), nn.Dropout(dropout)
        )

    def forward(self, x, context=None, mask=None, pos_emb=None):
        """

        Args:
            x: The 'latent array' in the Perceiver paper.
            context: The 'byte array' in the Perceiver paper (the input data).
            mask:
            pos_emb:

        Returns:

        """

        h = self.heads

        q = self.to_q(x)
        context = default(context, x)
        k, v = self.to_kv(context).chunk(2, dim=-1)

        # Rearrange the query, key and value tensors.
        # b = batch size; n = TODO (PD-2021-09-13)
        # h = number of heads; d = number of dims per head.
        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> (b h) n d", h=h), (q, k, v)
        )

        if exists(pos_emb):
            q, k = apply_rotary_emb(q, k, pos_emb)

        sim = einsum("b i d, b j d -> b i j", q, k) * self.scale

        if exists(mask):
            mask = rearrange(mask, "b ... -> b (...)")
            max_neg_value = -torch.finfo(sim.dtype).max
            mask = repeat(mask, "b j -> (b h) () j", h=h)
            sim.masked_fill_(~mask, max_neg_value)

        # attention, what we cannot get enough of
        attn = sim.softmax(dim=-1)

        out = einsum("b i j, b j d -> b i d", attn, v)
        out = rearrange(out, "(b h) n d -> b n (h d)", h=h)
        return self.to_out(out)



import torch
from torch import nn, einsum
import torch.nn.functional as F

from einops import rearrange, repeat

from perceiver_pytorch.layers import exists, default, cache_fn, PreNorm, FeedForward, Attention
from perceiver_pytorch.utils import fourier_encode


# helpers


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x, **kwargs):
        return x + self.fn(x, **kwargs)


class GRUGating(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.dim = dim
        self.fn = fn
        self.gru = nn.GRUCell(dim, dim)

    def forward(self, x, **kwargs):
        b, dim = x.shape[0], self.dim
        y = self.fn(x, **kwargs)

        gated_output = self.gru(
            rearrange(y, "... d -> (...) d"), rearrange(x, "... d -> (...) d")
        )

        gated_output = rearrange(gated_output, "(b n) d -> b n d", b=b)
        return gated_output


# main class


class Perceiver(nn.Module):
    def __init__(
        self,
        *,
        num_freq_bands,
        depth,
        max_freq,
        freq_base=2,
        input_channels=3,
        input_axis=2,
        num_latents=512,
        latent_dim=512,
        cross_heads=1,
        latent_heads=8,
        cross_dim_head=64,
        latent_dim_head=64,
        num_classes=1000,
        attn_dropout=0.0,
        ff_dropout=0.0,
        weight_tie_layers=False
    ):
        super().__init__()
        self.input_axis = input_axis
        self.max_freq = max_freq
        self.num_freq_bands = num_freq_bands
        self.freq_base = freq_base

        input_dim = input_axis * ((num_freq_bands * 2) + 1) + input_channels

        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))

        get_cross_attn = lambda: GRUGating(
            latent_dim,
            PreNorm(
                latent_dim,
                Attention(
                    latent_dim,
                    input_dim,
                    heads=cross_heads,
                    dim_head=cross_dim_head,
                    dropout=attn_dropout,
                ),
                context_dim=input_dim,
            ),
        )
        get_latent_attn = lambda: GRUGating(
            latent_dim,
            PreNorm(
                latent_dim,
                Attention(
                    latent_dim,
                    heads=latent_heads,
                    dim_head=latent_dim_head,
                    dropout=attn_dropout,
                ),
            ),
        )
        get_cross_ff = lambda: Residual(
            PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))
        )
        get_latent_ff = lambda: Residual(
            PreNorm(latent_dim, FeedForward(latent_dim, dropout=ff_dropout))
        )

        get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff = map(
            cache_fn,
            (get_cross_attn, get_cross_ff, get_latent_attn, get_latent_ff),
        )

        self.layers = nn.ModuleList([])
        for i in range(depth):
            should_cache = i > 0 and weight_tie_layers
            cache_args = {"_cache": should_cache}

            self.layers.append(
                nn.ModuleList(
                    [
                        get_cross_attn(**cache_args),
                        get_cross_ff(**cache_args),
                        get_latent_attn(**cache_args),
                        get_latent_ff(**cache_args),
                    ]
                )
            )

        self.to_logits = nn.Sequential(
            nn.LayerNorm(latent_dim), nn.Linear(latent_dim, num_classes)
        )

    def forward(self, data, mask=None):
        b, *axis, _, device = *data.shape, data.device
        assert (
            len(axis) == self.input_axis
        ), "input data must have the right number of axis"

        # calculate fourier encoded positions in the range of [-1, 1], for all axis

        axis_pos = list(
            map(
                lambda size: torch.linspace(
                    -1.0, 1.0, steps=size, device=device
                ),
                axis,
            )
        )
        pos = torch.stack(torch.meshgrid(*axis_pos), dim=-1)
        enc_pos = fourier_encode(
            pos, self.max_freq, self.num_freq_bands, base=self.freq_base
        )
        enc_pos = rearrange(enc_pos, "... n d -> ... (n d)")
        enc_pos = repeat(enc_pos, "... -> b ...", b=b)

        # concat to channels of data and flatten axis

        data = torch.cat((data, enc_pos), dim=-1)
        data = rearrange(data, "b ... d -> b (...) d")

        x = repeat(self.latents, "n d -> b n d", b=b)

        for cross_attn, cross_ff, latent_attn, latent_ff in self.layers:
            x = cross_attn(x, context=data, mask=mask)
            x = cross_ff(x)
            x = latent_attn(x)
            x = latent_ff(x)

        x = x.mean(dim=-2)
        return self.to_logits(x)



from math import log, pi

import torch
import torch.nn.functional as F
import numpy as np
import math
import einops


def extract_image_patches(
    x: torch.Tensor, kernel: int, stride: int = 1, dilation: int = 1
) -> torch.Tensor:
    """
    Extract image patches in a way similar to TensorFlow extract_image_patches
    Taken from https://discuss.pytorch.org/t/tf-extract-image-patches-in-pytorch/43837/8

    In the Perceiver JAX implementation they extract image patches matching TensorFlow's SAME padding.
    PyTorch doesn't have that same kind of option, so this is a way to do that.

    Args:
        x: Input Torch Tensor
        kernel: Size of kernel
        stride: Stride of patch
        dilation: Dilation rate

    Returns:
    Tensor of size [Batch, Height, Width, Channels*kernel*stride]

    """
    # Do TF 'SAME' Padding
    b, c, h, w = x.shape
    h2 = math.ceil(h / stride)
    w2 = math.ceil(w / stride)
    pad_row = (h2 - 1) * stride + (kernel - 1) * dilation + 1 - h
    pad_col = (w2 - 1) * stride + (kernel - 1) * dilation + 1 - w
    x = F.pad(x, (pad_row // 2, pad_row - pad_row // 2, pad_col // 2, pad_col - pad_col // 2))

    # Extract patches
    # get all image windows of size (kernel, stride) and stride (kernel, stride)
    patches = x.unfold(2, kernel, stride).unfold(3, kernel, stride)
    # Permute so that channels are next to patch dimension
    patches = patches.permute(0, 4, 5, 1, 2, 3).contiguous()
    # View as [batch_size, height, width, channels*kh*kw]
    return patches.view(b, -1, patches.shape[-2], patches.shape[-1])


def reverse_space_to_depth(
    frames: torch.Tensor, temporal_block_size: int = 1, spatial_block_size: int = 1
) -> torch.Tensor:
    """Reverse space to depth transform.
    Works for images (dim = 4) and videos (dim = 5)"""
    if len(frames.shape) == 4:
        return einops.rearrange(
            frames,
            "b (dh dw c) h w -> b c (h dh) (w dw)",
            dh=spatial_block_size,
            dw=spatial_block_size,
        )
    elif len(frames.shape) == 5:
        return einops.rearrange(
            frames,
            "b t (dt dh dw c) h w -> b (t dt) c (h dh) (w dw)",
            dt=temporal_block_size,
            dh=spatial_block_size,
            dw=spatial_block_size,
        )
    else:
        raise ValueError(
            "Frames should be of rank 4 (batch, height, width, channels)"
            " or rank 5 (batch, time, height, width, channels)"
        )


def space_to_depth(
    frames: torch.Tensor, temporal_block_size: int = 1, spatial_block_size: int = 1
) -> torch.Tensor:
    """Space to depth transform.
    Works for images (dim = 4) and videos (dim = 5)"""
    if len(frames.shape) == 4:
        return einops.rearrange(
            frames,
            "b c (h dh) (w dw) -> b (dh dw c) h w",
            dh=spatial_block_size,
            dw=spatial_block_size,
        )
    elif len(frames.shape) == 5:
        return einops.rearrange(
            frames,
            "b (t dt) c (h dh) (w dw) -> b t (dt dh dw c) h w ",
            dt=temporal_block_size,
            dh=spatial_block_size,
            dw=spatial_block_size,
        )
    else:
        raise ValueError(
            "Frames should be of rank 4 (batch, height, width, channels)"
            " or rank 5 (batch, time, height, width, channels)"
        )


def encode_position(
    batch_size: int,
    axis: list,
    max_frequency: float,
    num_frequency_bands: int,
    sine_only: bool = False,
) -> torch.Tensor:
    """
    Encode the Fourier Features and return them

    Args:
        batch_size: Batch size
        axis: List containing the size of each axis
        max_frequency: Max frequency
        num_frequency_bands: Number of frequency bands to use
        sine_only: (bool) Whether to only use Sine features or both Sine and Cosine, defaults to both

    Returns:
        Torch tensor containing the Fourier Features of shape [Batch, *axis]
    """
    axis_pos = list(
        map(
            lambda size: torch.linspace(-1.0, 1.0, steps=size),
            axis,
        )
    )
    pos = torch.stack(torch.meshgrid(*axis_pos), dim=-1)
    enc_pos = fourier_encode(
        pos,
        max_frequency,
        num_frequency_bands,
        sine_only=sine_only,
    )
    enc_pos = einops.rearrange(enc_pos, "... n d -> ... (n d)")
    enc_pos = einops.repeat(enc_pos, "... -> b ...", b=batch_size)
    return enc_pos


def fourier_encode(
    x: torch.Tensor,
    max_freq: float,
    num_bands: int = 4,
    sine_only: bool = False,
) -> torch.Tensor:
    """
    Create Fourier Encoding

    Args:
        x: Input Torch Tensor
        max_freq: Maximum frequency for the Fourier features
        num_bands: Number of frequency bands
        sine_only: Whether to only use sine or both sine and cosine features

    Returns:
        Torch Tensor with the fourier position encoded concatenated
    """
    x = x.unsqueeze(-1)
    device, dtype, orig_x = x.device, x.dtype, x

    scales = torch.linspace(
        1.0,
        max_freq / 2,
        num_bands,
        device=device,
        dtype=dtype,
    )
    scales = scales[(*((None,) * (len(x.shape) - 1)), Ellipsis)]

    x = x * scales * pi
    x = x.sin() if sine_only else torch.cat([x.sin(), x.cos()], dim=-1)
    x = torch.cat((x, orig_x), dim=-1)
    return x



import torch
from torch import nn, einsum
from einops import rearrange, repeat


class SinusoidalEmbeddings(nn.Module):
    def __init__(self, dim):
        super().__init__()
        inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

    def forward(self, x):
        n = x.shape[-2]
        t = torch.arange(n, device=x.device).type_as(self.inv_freq)
        sinusoid_inp = torch.einsum("i , j -> i j", t, self.inv_freq)
        emb = torch.cat((sinusoid_inp.sin(), sinusoid_inp.cos()), dim=-1)
        return emb[None, :, :]


def rotate_every_two(x):
    x = rearrange(x, "... (d j) -> ... d j", j=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, "... d j -> ... (d j)")


def apply_rotary_emb(q, k, sinu_pos):
    sinu_pos = rearrange(sinu_pos, "() n (j d) -> n j d", j=2)
    sin, cos = sinu_pos.unbind(dim=-2)
    sin, cos = map(lambda t: repeat(t, "b n -> b (n j)", j=2), (sin, cos))
    q, k = map(lambda t: (t * cos) + (rotate_every_two(t) * sin), (q, k))
    return q, k



import torch
from torch import nn

from einops import repeat
from perceiver_pytorch.layers import exists, cache_fn, PreNorm, FeedForward, Attention

# main class

class PerceiverIO(nn.Module):
    def __init__(
        self,
        *,
        depth,
        dim,
        queries_dim,
        logits_dim = None,
        num_latents = 512,
        latent_dim = 512,
        cross_heads = 1,
        latent_heads = 8,
        cross_dim_head = 64,
        latent_dim_head = 64,
        weight_tie_layers = False,
        decoder_ff = False
    ):
        """
        PerceiverIO implementation from https://arxiv.org/abs/2107.14795

        Args:
            depth: Depth of the network
            dim: dimension of sequence to be encoded
            queries_dim: Dimension of the decoder queries
            logits_dim: Dimension of final logits
            num_latents: Number of latents
            latent_dim: Latent dimension
            cross_heads: Number of heads for cross attention
            latent_heads: Number of heads for latent self-attention
            cross_dim_head: Number of dimensions per cross attention head
            latent_dim_head: Number of dimensions per latent self-attention head
            weight_tie_layers: Whether to weight tie layers
            decoder_ff: Whether to use a feed forward network on the decoder queries
        """
        super().__init__()
        self.latents = nn.Parameter(torch.randn(num_latents, latent_dim))

        self.cross_attend_blocks = nn.ModuleList([
            PreNorm(latent_dim, Attention(latent_dim, dim, heads = cross_heads, dim_head = cross_dim_head), context_dim = dim),
            PreNorm(latent_dim, FeedForward(latent_dim))
        ])

        get_latent_attn = lambda: PreNorm(latent_dim, Attention(latent_dim, heads = latent_heads, dim_head = latent_dim_head))
        get_latent_ff = lambda: PreNorm(latent_dim, FeedForward(latent_dim))
        get_latent_attn, get_latent_ff = map(cache_fn, (get_latent_attn, get_latent_ff))

        self.layers = nn.ModuleList([])
        cache_args = {'_cache': weight_tie_layers}

        for i in range(depth):
            self.layers.append(nn.ModuleList([
                get_latent_attn(**cache_args),
                get_latent_ff(**cache_args)
            ]))

        self.decoder_cross_attn = PreNorm(queries_dim, Attention(queries_dim, latent_dim, heads = cross_heads, dim_head = cross_dim_head), context_dim = latent_dim)
        self.decoder_ff = PreNorm(queries_dim, FeedForward(queries_dim)) if decoder_ff else None

        self.to_logits = nn.Linear(queries_dim, logits_dim) if exists(logits_dim) else nn.Identity()

    def forward(
        self,
        data,
        mask = None,
        queries = None
    ):
        b, *_, device = *data.shape, data.device

        x = repeat(self.latents, 'n d -> b n d', b = b)

        cross_attn, cross_ff = self.cross_attend_blocks

        # cross attention only happens once for Perceiver IO

        x = cross_attn(x, context = data, mask = mask) + x
        x = cross_ff(x) + x

        # layers

        for self_attn, self_ff in self.layers:
            x = self_attn(x) + x
            x = self_ff(x) + x

        if not exists(queries):
            return x

        # cross attend from decoder queries to latents
        
        latents = self.decoder_cross_attn(queries, context = x)

        # optional decoder feedforward

        if exists(self.decoder_ff):
            latents = latents + self.decoder_ff(latents)

        # final linear out

        return self.to_logits(latents)




import torch
from torch.nn import functional as F


class Conv2DDownsample(torch.nn.Sequential):
    def __init__(
        self,
        num_layers: int = 1,
        input_channels: int = 12,
        output_channels: int = 64,
        use_batchnorm: bool = True,
    ):
        """
        Constructs a Conv2DDownsample model
        Args:
            num_layers: Number of conv -> maxpool layers
            output_channels: Number of output channels
            input_channels: Number of input channels to first layer
            use_batchnorm: Whether to use Batch Norm
        """

        layers = [self.make_layer(input_channels, output_channels, batch=use_batchnorm)]
        for _ in range(num_layers - 1):
            layers += [
                self.make_layer(output_channels, output_channels, batch=use_batchnorm)
            ]

        super().__init__(*layers)

    def make_layer(self, c_in, c_out, ks=7, stride=2, padding=7 // 2, batch=True):
        "Make Conv->Batch->Relu->MaxPool stack"
        layers = [torch.nn.Conv2d(c_in, c_out, ks, stride, padding, bias=False)]
        if batch:
            layers += [torch.nn.BatchNorm2d(c_out)]
        layers += [torch.nn.ReLU(), torch.nn.MaxPool2d(3, stride=2, padding=3 // 2)]
        return torch.nn.Sequential(*layers)


class Conv2DUpsample(torch.nn.Module):
    def __init__(self, input_channels: int = 12, output_channels: int = 12):
        """
        Upsamples 4x using 2 2D transposed convolutions
        Args:
            input_channels: Input channels to the first layer
            output_channels: Number of output channels
        """

        super().__init__()
        self.transpose_conv1 = torch.nn.ConvTranspose2d(
            in_channels=input_channels,
            out_channels=output_channels * 2,
            kernel_size=(4, 4),
            stride=(2, 2),
            padding=(1, 1),
        )
        self.transpose_conv2 = torch.nn.ConvTranspose2d(
            in_channels=output_channels * 2,
            out_channels=output_channels,
            kernel_size=(4, 4),
            stride=(2, 2),
            padding=(1, 1),
        )

    def forward(self, x):
        x = self.transpose_conv1(x)
        x = F.relu(x)
        x = self.transpose_conv2(x)
        return x


class Conv3DUpsample(torch.nn.Module):
    def __init__(
        self,
        input_channels: int = 12,
        output_channels: int = 12,
        num_temporal_upsamples: int = 2,
        num_space_upsamples: int = 4,
    ):
        """
        Simple convolutional auto-encoder
        Args:
            output_channels: Final output channels
            num_temporal_upsamples: Number of temporal upsamples to perform
            num_space_upsamples: Number of spatial upsamples to perform
        """

        super().__init__()
        temporal_stride = 2
        space_stride = 2
        num_upsamples = max(num_space_upsamples, num_temporal_upsamples)

        # create the input and output changesl for the different layers
        # The intermediate channels are (output channels) * 2^(number of upsamples - 1 - index of the current upsample)
        # The decoder sets the number of upsamples as log2(upsample_value), and this changes the number of channels
        # in a similar way, so it all scales together.
        intermediate_output_channels = [
            output_channels * pow(2, num_upsamples - 1 - i)
            for i in range(0, num_upsamples)
        ]
        intermediate_input_channels = [input_channels] + intermediate_output_channels

        self.layers = torch.nn.ModuleList()
        for i in range(num_upsamples):
            if i >= num_temporal_upsamples:
                temporal_stride = 1
            if i >= num_space_upsamples:
                space_stride = 1

            input_channels = input_channels if i == 0 else output_channels
            stride = (temporal_stride, space_stride, space_stride)
            conv = torch.nn.ConvTranspose3d(
                in_channels=intermediate_input_channels[i],
                out_channels=intermediate_output_channels[i],
                stride=stride,
                kernel_size=stride,
            )
            # see output dims calculations - https://pytorch.org/docs/stable/generated/torch.nn.ConvTranspose3d.html
            # if kernel=stride, and dilation -1
            # D = (D-1) * stride + (kernel - 1) + 2 = D * stride

            self.layers.append(conv)
            if i != num_upsamples - 1:
                self.layers.append(torch.nn.ReLU())

    def forward(self, x):
        # Move around to Channels first
        x = x.permute(0, 2, 1, 3, 4)

        # loop over layers
        for layer in self.layers:
            x = layer(x)

        x = F.relu(x)
        # Move back
        x = x.permute(0, 2, 1, 3, 4)
        return x



import torch
import numpy as np
from perceiver_pytorch.utils import reverse_space_to_depth
from perceiver_pytorch.convolutions import Conv2DUpsample, Conv3DUpsample


class ImageDecoder(torch.nn.Module):
    def __init__(
        self,
        postprocess_type: str = "pixels",
        spatial_upsample: int = 1,
        temporal_upsample: int = 1,
        output_channels: int = -1,
        input_channels: int = 12,
        input_reshape_size=None,
    ):
        """
        ImageDecoder modeled after JAX version here
        https://github.com/deepmind/deepmind-research/blob/769bfdbeafbcb472cb8e2c6cfa746b53ac82efc2/perceiver/io_processors.py#L441-L510

        Args:
            postprocess_type: Type of postprocessing, one of conv, patches, pixels, raft, or conv1x1
            spatial_upsample: How much to spatially upsample
            temporal_upsample: How much to temporally upsample
            output_channels: Number of output channels, should be the final desired number of channels
                Has to explicitly set for conv and conv1x1 options, otherwise an error will be raised.
                Ignored for patches and pixels options.
            input_channels: Number of input channels to decoder
            input_reshape_size: The size to reshape the input to
        """

        super().__init__()

        if postprocess_type not in ("conv", "patches", "pixels", "conv1x1"):
            # TODO Add Raft
            raise ValueError("Invalid postprocess_type!")

        # Architecture parameters:
        self.postprocess_type = postprocess_type

        self.temporal_upsample = temporal_upsample
        self.spatial_upsample = spatial_upsample
        self.input_reshape_size = input_reshape_size

        if postprocess_type == "pixels":
            # No postprocessing for pixels
            self.decoder = torch.nn.Identity()
        elif postprocess_type == "patches":
            self.decoder = ImageDecoderPatches(
                spatial_upsample=spatial_upsample, temporal_upsample=temporal_upsample
            )
        elif postprocess_type == "conv":
            self.decoder = ImageDecoderConv(
                spatial_upsample=spatial_upsample,
                temporal_upsample=temporal_upsample,
                output_channels=output_channels,
                input_channels=input_channels,
            )
        elif postprocess_type == "conv1x1":
            self.decoder = ImageDecoderConv1x1(
                spatial_upsample=spatial_upsample,
                output_channels=output_channels,
                input_channels=input_channels,
            )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        if self.input_reshape_size is not None:
            inputs = torch.reshape(
                inputs,
                [inputs.shape[0]] + list(self.input_reshape_size) + [inputs.shape[-1]],
            )
        return self.decoder(inputs)


class ImageDecoderConv(torch.nn.Module):
    def __init__(
        self,
        spatial_upsample: int = 1,
        temporal_upsample: int = 1,
        output_channels: int = -1,
        input_channels: int = 12,
    ):
        """
        Convolutional image decoder that can upsample temporally and spatially

        Args:
            spatial_upsample: How much to spatially upsample
            temporal_upsample: How much to temporally upsample
            output_channels: Number of output channels, should be the final desired number of channels
                Has to explicitly set for conv and conv1x1 options, otherwise an error will be raised.
                Ignored for patches and pixels options.
            input_channels: Number of input channels to decoder
        """

        super().__init__()

        self.temporal_upsample = temporal_upsample
        self.spatial_upsample = spatial_upsample

        if output_channels == -1:
            raise ValueError("Expected value for output_channels")
        if self.temporal_upsample != 1:

            def int_log2(x):
                return int(np.round(np.log(x) / np.log(2)))

            self.convnet = Conv3DUpsample(
                input_channels=input_channels,
                output_channels=output_channels,
                num_temporal_upsamples=int_log2(temporal_upsample),
                num_space_upsamples=int_log2(spatial_upsample),
            )
        else:
            assert (
                self.spatial_upsample == 4
            ), "Conv2DUpsample only support 4x spatial upsample right now"
            self.convnet = Conv2DUpsample(
                input_channels=input_channels, output_channels=output_channels
            )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        # Convnet image featurization.
        if len(inputs.shape) == 5 and self.temporal_upsample == 1:
            # Timeseries, do it to each timestep independently
            outs = []
            for i in range(inputs.shape[1]):
                outs.append(self.convnet(inputs[:, i, :, :, :]))
            inputs = torch.stack(outs, dim=1)
        else:
            inputs = self.convnet(inputs)

        return inputs


class ImageDecoderConv1x1(torch.nn.Module):
    def __init__(
        self,
        spatial_upsample: int = 1,
        output_channels: int = -1,
        input_channels: int = 12,
    ):
        """
        Convolutional 1x1 image decoder

        Args:
            spatial_upsample: How much to spatially upsample
            output_channels: Number of output channels, should be the final desired number of channels
                Has to explicitly set for conv and conv1x1 options, otherwise an error will be raised.
                Ignored for patches and pixels options.
            input_channels: Number of input channels to decoder
        """

        super().__init__()

        self.spatial_upsample = spatial_upsample

        if output_channels == -1:
            raise ValueError("Expected value for output_channels")
        self.conv1x1 = torch.nn.Conv2d(
            in_channels=input_channels,
            out_channels=output_channels,
            kernel_size=(1, 1),
            # spatial_downsample is unconstrained for 1x1 convolutions.
            stride=(self.spatial_upsample, self.spatial_upsample),
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        # Convnet image featurization.
        if len(inputs.shape) == 5:
            # Timeseries, do it to each timestep independently
            outs = []
            for i in range(inputs.shape[1]):
                outs.append(self.conv1x1(inputs[:, i, :, :, :]))
            inputs = torch.stack(outs, dim=1)
        else:
            inputs = self.conv1x1(inputs)

        return inputs


class ImageDecoderPatches(torch.nn.Module):
    def __init__(
        self, spatial_upsample: int = 1, temporal_upsample: int = 1,
    ):
        """
        Patch-based image decoder

        Args:
            spatial_upsample: How much to spatially upsample
            temporal_upsample: How much to temporally upsample
        """

        super().__init__()

        self.temporal_upsample = temporal_upsample
        self.spatial_upsample = spatial_upsample

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        inputs = reverse_space_to_depth(
            inputs, self.temporal_upsample, self.spatial_upsample
        )
        return inputs



import torch
from torch import nn
import torchvision
import torch.nn.functional as F
import numpy as np
import math

from perceiver_pytorch.convolutions import Conv2DDownsample
from perceiver_pytorch.utils import space_to_depth


class ImageEncoder(torch.nn.Module):
    def __init__(
        self,
        input_channels: int = 12,
        prep_type: str = "conv",
        spatial_downsample: int = 4,
        temporal_downsample: int = 1,
        output_channels: int = 64,
        conv2d_use_batchnorm: bool = True,
        crop_size: int = 256,
        use_space2depth: bool = True,
    ):
        """
        Image encoder class, modeled off the JAX version
        https://github.com/deepmind/deepmind-research/blob/769bfdbeafbcb472cb8e2c6cfa746b53ac82efc2/perceiver/io_processors.py#L291-L438

        Args:
            input_channels: Number of input channels of the original image/video
            prep_type: How to encode the images, one of conv, patches, pixels, or conv1x1
            spatial_downsample: How much to downsample spatially
            temporal_downsample: How much to downsample temporally
            output_channels: Number of output channels to send to Perceiver
            conv2d_use_batchnorm: Whether to use batch norm
            crop_size: Only for MetNet preprocessor, the center crop size
            use_space2depth: Only for MetNet preprocessor, whether to use average pooling, or space2depth for downsampling
        """
        super().__init__()
        self.prep_type = prep_type

        if prep_type not in ("conv", "patches", "pixels", "conv1x1", "metnet"):
            raise ValueError("Invalid prep_type!")

        if self.prep_type == "conv":
            self.encoder = ImageEncoderConv(
                input_channels=input_channels,
                temporal_downsample=temporal_downsample,
                spatial_downsample=spatial_downsample,
                output_channels=output_channels,
                conv2d_use_batchnorm=conv2d_use_batchnorm,
            )
        elif self.prep_type == "conv1x1":
            self.encoder = ImageEncoderConv1x1(
                input_channels=input_channels,
                spatial_downsample=spatial_downsample,
                output_channels=output_channels,
            )
        elif self.prep_type == "patches":
            self.encoder = ImageEncoderPatches(
                temporal_downsample=temporal_downsample,
                spatial_downsample=spatial_downsample,
            )
        elif self.prep_type == "pixels":
            self.encoder = ImageEncoderPixel(
                temporal_downsample=temporal_downsample,
                spatial_downsample=spatial_downsample,
            )
        elif self.prep_type == "metnet":
            self.encoder = ImageEncoderMetNet(
                crop_size=crop_size, use_space2depth=use_space2depth
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.encoder(x)


class ImageEncoderConv(torch.nn.Module):
    def __init__(
        self,
        input_channels: int = 12,
        spatial_downsample: int = 4,
        temporal_downsample: int = 1,
        output_channels: int = 64,
        conv2d_use_batchnorm: bool = True,
    ):
        """
        Convolutional image encoder that can spatially and temporally downsample

        Args:
            input_channels: Number of input channels of the original image/video
            spatial_downsample: How much to downsample spatially
            temporal_downsample: How much to downsample temporally
            output_channels: Number of output channels to send to Perceiver
            conv2d_use_batchnorm: Whether to use batch norm
        """
        super().__init__()
        self.temporal_downsample = temporal_downsample
        self.spatial_downsample = spatial_downsample
        self.output_channels = output_channels

        # Downsampling with conv is currently restricted
        convnet_num_layers = math.log(spatial_downsample, 4)
        convnet_num_layers_is_int = convnet_num_layers == np.round(convnet_num_layers)
        if not convnet_num_layers_is_int or temporal_downsample != 1:
            raise ValueError(
                "Only powers of 4 expected for spatial "
                "and 1 expected for temporal "
                "downsampling with conv."
            )

        self.convnet = Conv2DDownsample(
            num_layers=int(convnet_num_layers),
            output_channels=output_channels,
            input_channels=input_channels,
            use_batchnorm=conv2d_use_batchnorm,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 5:
            # Timeseries, do it to each timestep independently
            outs = []
            for i in range(x.shape[1]):
                outs.append(self.convnet(x[:, i, :, :, :]))
            x = torch.stack(outs, dim=1)
        else:
            x = self.convnet(x)
        return x


class ImageEncoderConv1x1(torch.nn.Module):
    def __init__(
        self,
        input_channels: int = 12,
        spatial_downsample: int = 4,
        output_channels: int = 64,
    ):
        """
        Convolutional 1x1 encoder that can spatially downsample

        Args:
            input_channels: Number of input channels of the original image/video
            spatial_downsample: How much to downsample spatially
            output_channels: Number of output channels to send to Perceiver
        """
        super().__init__()
        self.spatial_downsample = spatial_downsample
        self.output_channels = output_channels

        self.convnet_1x1 = torch.nn.Conv2d(
            in_channels=input_channels,
            out_channels=output_channels,
            kernel_size=(1, 1),
            # spatial_downsample is unconstrained for 1x1 convolutions.
            stride=(spatial_downsample, spatial_downsample),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if len(x.shape) == 5:
            # Timeseries, do it to each timestep independently
            outs = []
            for i in range(x.shape[1]):
                outs.append(self.convnet_1x1(x[:, i, :, :, :]))
            x = torch.stack(outs, dim=1)
        else:
            x = self.convnet_1x1(x)

        return x


class ImageEncoderPatches(torch.nn.Module):
    def __init__(
        self, spatial_downsample: int = 4, temporal_downsample: int = 1,
    ):
        """
        Image encoder that uses patches

        Args:
            spatial_downsample: How much to downsample spatially
            temporal_downsample: How much to downsample temporally
        """
        super().__init__()
        self.temporal_downsample = temporal_downsample
        self.spatial_downsample = spatial_downsample

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = space_to_depth(
            x,
            temporal_block_size=self.temporal_downsample,
            spatial_block_size=self.spatial_downsample,
        )

        # For flow
        if x.ndim == 5 and x.shape[1] == 1:
            x = x.squeeze(axis=1)

        return x


class ImageEncoderPixel(torch.nn.Module):
    def __init__(
        self, spatial_downsample: int = 4, temporal_downsample: int = 1,
    ):
        """
        Image encoder class for simple downsampling with pixels

        Args:
            spatial_downsample: How much to downsample spatially
            temporal_downsample: How much to downsample temporally
        """
        super().__init__()
        self.temporal_downsample = temporal_downsample
        self.spatial_downsample = spatial_downsample

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        # If requested, will downsample in simplest way
        if x.ndim == 4:
            x = x[:, :, :: self.spatial_downsample, :: self.spatial_downsample]
        elif x.ndim == 5:
            x = x[
                :,
                :: self.temporal_downsample,
                :,
                :: self.spatial_downsample,
                :: self.spatial_downsample,
            ]
        else:
            raise ValueError("Unsupported data format for pixels")

        return x


class ImageEncoderMetNet(nn.Module):
    def __init__(
        self, crop_size: int = 256, use_space2depth: bool = True,
    ):
        """
        Performs the MetNet preprocessing of mean pooling Sat channels, followed by
        concatenating the center crop and mean pool

        In the paper, the radar data is space2depth'd, while satellite channel is mean pooled, but for this different
        task, we choose to do either option for satellites

        Args:
            sat_channels: Number of satellite channels
            crop_size: Center crop size
            use_space2depth: Whether to use space2depth on satellite channels, or mean pooling, like in paper
        """
        super().__init__()
        # Split off sat + mask channels into own image, and the rest, which we just take a center crop
        # For this,
        self.sat_downsample = (
            torch.nn.PixelUnshuffle(downscale_factor=2)
            if use_space2depth
            else torch.nn.AvgPool3d(kernel_size=(1, 2, 2))
        )
        self.center_crop = torchvision.transforms.CenterCrop(size=crop_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.sat_downsample(x)
        # In paper, satellite and radar data is concatenated here
        # We are just going to skip that bit
        sat_center = self.center_crop(x)
        sat_mean = F.avg_pool3d(x, (1, 2, 2))
        # All the same size now, so concatenate together, already have time, lat/long, and elevation image
        x = torch.cat([sat_center, sat_mean], dim=2)
        return x
